/*

DS glOrtho Test
relminator 04/07/2010 
Richard Eric M. Lope BSN RN

http://rel.betterwebber.com

Just a proof of concept of using the 3d core to render 2d stuff
I'm planning to use this on my game space impakto DS
maybe for bullets or enemies as I have no Idea how to render oam-based
sprites in stencil mode. (ie. Render an outline of the sprite in a single color)


Notes:
	1. Matrix is column-major order
	2. Draworder is from last GFX_VERTEX call to first GFX_VERTEX call ie: reversed painter's algo
		Hence: the g_depth++ for every raster.
	3. Vertex calls are have absolute coordinates (unlike OpenGL's relative coords)
	4. No GL_DEPTH_TEST!!!!! Hence: #2
	
To do:
	1. Textured quads as sprites(done)
	2. Load spriteset funk
	3. Alpha blending(dunno how yet)
	4. Rotation (Use the 2d core instead?)
	5. encapsulate into a better class so that g_depth, textures, spritesizes, etc
		can be managed easily.

*/


#include <nds.h>
#include <stdio.h>

#include "texture_2x2_16.h"

// macros for easy referencing

#define glVertex2i(x,y) ({ GFX_VERTEX16 = (((u16)(y)) << 16) | ((u16)(x)); GFX_VERTEX16 = g_depth; })	
#define glVertex3i(x,y,z) ({ GFX_VERTEX16 = (((u16)(y)) << 16) | ((u16)(x)); GFX_VERTEX16 = z; })
#define glTexcoord2i(u,v) ({ GFX_TEX_COORD = ((u16)(v)) << 20 | ((u16)(u)) << 4; })	



// Needs this since the 3d core renders stuff in reverse order for vertices with
// the same z-value
s16 g_depth = 0;

// yay!! global textures. yuck!
int g_textureID;  


// Reset TEXIMAGE_PARAM (40004A8h - Cmd 2Ah)
// What's the libnds equivalent?
// Needs to be called so that primitives does not
// get funky colors from texture

inline void gl2d_disable_texture() 
{
	GFX_TEX_FORMAT = 0;
}


// initializes gl2d
void gl2d_init()
{
		
	glInit();
	glViewport(0,0,255,191);
	
	glClearColor(0,0,0,31);
	glClearDepth(GL_MAX_DEPTH);
	

}

// set orthographic projection
// at 1:1 correspondence to screen coords
void gl2d_set_ortho()		
{

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrthof32(0, SCREEN_WIDTH, SCREEN_HEIGHT, 0, -4097, 1);

}


// issue this before drawing 2d
void gl2d_start_2dmode()
{

	// what?!! No glDisable(GL_DEPTH_TEST)?!!!!!!
	
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_ANTIALIAS);
	glDisable(GL_OUTLINE);

	glColor(0x7FFF); 
	
	glPolyFmt(POLY_ALPHA(31) | POLY_CULL_NONE);


	glPushMatrix();
	gl2d_set_ortho();
	
	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// set depth to 0
	g_depth = 0;

}

// issue this after drawing 2d so that we don't mess the matrix stack
void gl2d_end_2dmode()
{
	glPopMatrix(1);
}


// draws a filled box
void gl2d_box_fill(s32 x1, s32 y1, s32 x2, s32 y2, s32 color)
{
	
	glColor(color);
	glBegin(GL_QUADS);
		glVertex2i(x1,y1);
		glVertex2i(x2,y1);
		glVertex2i(x2,y2);
		glVertex2i(x1,y2);
	glEnd();
	
	g_depth++;
	
}

// draws a line
void gl2d_line(s32 x1, s32 y1, s32 x2, s32 y2, s32 color)
{

	glColor(color);
	glBegin(GL_TRIANGLES);
		glVertex2i(x1,y1);
		glVertex2i(x2,y2);
		glVertex2i(x2,y2);
	glEnd();
	
	g_depth++;

}

// draws a box
void gl2d_box(s32 x1, s32 y1, s32 x2, s32 y2, s32 color)
{
	
	glColor(color);
	glBegin(GL_TRIANGLES);
	
		glVertex2i(x1,y1);
		glVertex2i(x2,y1);
		glVertex2i(x2,y1);

		glVertex2i(x2,y1);
		glVertex2i(x2,y2);
		glVertex2i(x2,y2);

		glVertex2i(x2,y2);
		glVertex2i(x1,y2);
		glVertex2i(x1,y2);

		glVertex2i(x1,y2);
		glVertex2i(x1,y1);
		glVertex2i(x1,y1);
		
	glEnd();
	
	g_depth++;
	
}


// temporary example of ortho sprites
void gl2d_sprite(s32 x, s32 y)
{
	int x1 = x;
	int y1 = y;
	int x2 = x+127;
	int y2 = y+127;
	
	
	glBindTexture(0, g_textureID);
	
	glBegin(GL_QUADS);
		glTexcoord2i(0,0);
		glVertex2i(x1,y1);
		
		glTexcoord2i(127,0);
		glVertex2i(x2,y1);
		
		glTexcoord2i(127,127);
		glVertex2i(x2,y2);
		
		glTexcoord2i(0,127);
		glVertex2i(x1,y2);
	glEnd();
	
	g_depth++;

}

// another temp sprite funk
// my sprite foo is strong!
void gl2d_sprite_foo(s32 x, s32 y, s32 u_off, s32 v_off)
{
	int x1 = x;
	int y1 = y;
	int x2 = x+63;
	int y2 = y+63;
	
	
	glBindTexture(0, g_textureID);
	
	glBegin(GL_QUADS);
		glTexcoord2i(u_off,v_off);
		glVertex2i(x1,y1);
		
		glTexcoord2i(u_off+63,v_off);
		glVertex2i(x2,y1);
		
		glTexcoord2i(u_off+63,v_off+63);
		glVertex2i(x2,y2);
		
		glTexcoord2i(u_off,v_off+63);
		glVertex2i(x1,y2);
	glEnd();
	
	g_depth++;

}


int main()
{

	videoSetMode(MODE_5_3D);
	
	
	consoleDemoInit();
	
	iprintf("\x1b[1;1HDS Ortho Test");
	iprintf("\x1b[2;1HUsing the 3d core");
	iprintf("\x1b[3;1HTo draw 2d primitives.");
		
	iprintf("\x1b[6;1HRelminator");
	iprintf("\x1b[7;1Hhttp://rel.betterwebber.com");
	
	
	gl2d_init();
	
	vramSetBankA(VRAM_A_TEXTURE);
	glGenTextures(1, &g_textureID);
	glBindTexture(0, g_textureID);
	glTexImage2D( 0, 0, GL_RGB, TEXTURE_SIZE_128 , TEXTURE_SIZE_128, 0, GL_TEXTURE_WRAP_S|GL_TEXTURE_WRAP_T|TEXGEN_TEXCOORD, (u8*)texture_2x2_16Bitmap );

	
	
	s32 x = 10;
	s32 y = 10;
	
	s32 x1,y1,x2,y2;
	u8 r,g,b;
	
	s32 frame = 0;
	while(1)
	{
	
		frame = (frame + 128) & 32767;
		x1 = rand() & 255;
		y1 = rand() & 255;
		x2 = rand() & 255;
		y2 = rand() & 255;
		
		r = rand() & 31;
		g = rand() & 31;
		b = rand() & 31;
		
		gl2d_start_2dmode();
	
			// needs to be called before texturing
			// or sprite gets blended by last glcolor call
			glColor(ARGB16(1,31,31,31)); 
	
			gl2d_sprite(10, 10);
			
			gl2d_disable_texture();
			
			x = 90;
			y = 90;
			
			gl2d_box_fill(x, y, x + 60, y + 40, ARGB16(1,0,31,31));
			
			x = 128;
			y = 100;
			
			gl2d_box_fill(x, y, x + 10, y+ 80, ARGB16(1,31,31,31));
			
			x = 10;
			y = 120;
			
			gl2d_box_fill(x, y, x + 200, y+ 30, ARGB16(1,31,20,3));
			
			
			
			gl2d_line(x1,y1,x2,y2,ARGB16(1,r,g,b));
			gl2d_box(x1,y1,x2,y2,ARGB16(1,r,g,b));
			
			
			// do some funky frame based movement
			x = 128 + ((cosLerp(frame)+sinLerp(32767/2+frame) * 70) >> 12);
			y = 96 + ((cosLerp(frame)+cosLerp(-frame) * 50) >> 12);
			
			// needs to be called before texturing
			// or sprite gets blended by last glcolor call
			glColor(ARGB16(1,31,31,31)); 
			
			gl2d_sprite_foo(x, 			y, 			64, 	64);
			gl2d_sprite_foo(255-x-63, 	191-y-63, 	0, 		64);
			gl2d_sprite_foo(255-x-63, 	y, 			64, 	0 );
			gl2d_sprite_foo(x, 			191-y-63, 	0, 		0 );
			
			
			
		 
		gl2d_end_2dmode();
		
		
		glFlush(0);

		swiWaitForVBlank();
	}

	return 0;
	
}
